# redis的密码,如果没有密码就是''
PASSWORD = ''
HOST = 'localhost'
PORT = '6379'
# 代理池的名称
PROXYPOOL = 'proxies'
# 测试网站
TEST_API = 'https://www.baidu.com/'
TEST_HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.106 Safari/537.36',
}

# 循环校验时间
CYCLE_VAILD_TIME = 60

# 代理池数量的最小值
LOWER_NUM = 60
# 代理池数量的最大值
UPPER_NUM = 600
# 循环检查添加时间
CHECK_POOL_CYCLE = 60



from flask import Flask, g
from proxy_pool.db import Reidis_client

__all__ = ['app']
app = Flask(__name__)


def get_conn():
    if not hasattr(g, 'redis_client'):
        g.redis_client = Reidis_client()
    return g.redis_client


@app.route('/')
def index():
    return '<h1>欢迎进入代理池系统！</h1>'


@app.route('/get')
def get():
    return get_conn().pop()


@app.route('/count')
def count():
    return str(get_conn().queue_len)



'''
db组件，主要用操作redis
'''
import redis
from proxy_pool.settings import *


class Reidis_client(object):
    def __init__(self):
        """
        考虑到：是否有密码
        """
        if PASSWORD:
            self.__db = redis.Redis(host=HOST, port=PORT, password=PASSWORD)
        else:
            self.__db = redis.Redis(host=HOST, port=PORT)

    def put(self, proxy):
        """
        添加一个代理到代理池
        :return:
        """
        self.__db.rpush(PROXYPOOL, proxy)

    def get(self, count=1):
        '''
        获取多个count个代理
        :return:
        '''
        proxies = self.__db.lrange(PROXYPOOL, 0, count - 1)
        self.__db.ltrim(PROXYPOOL, count, -1)
        return proxies

    def pop(self):
        # redis中存储的是bytes，用的时候必须是字符串。
        return self.__db.rpop(PROXYPOOL).decode('utf-8')

    @property
    def queue_len(self):
        return self.__db.llen(PROXYPOOL)


if __name__ == '__main__':
    redis_client = Reidis_client()
    print(redis_client.pop())



import time
from multiprocessing import Process
import asyncio
import aiohttp
from proxy_pool.db import Reidis_client
from proxy_pool.settings import *
from proxy_pool.getter import FreeProxyGetter

'''
校验器：
肯定要用db组件
'''


class VailidtyTester(object):

    def __init__(self):
        self.__raw_proxies = []
        # self.__conn = Reidis_client()

    # 这个方法一旦被调用，校验器要工作了
    def set_raw_proxies(self, proxies):
        self.__raw_proxies = proxies
        self.__conn = Reidis_client()

    # 校验代理的逻辑
    # 异步模块：aiohttp和asyncio
    async def test_singer_proxy(self, proxy):

        try:
            async with aiohttp.ClientSession() as session:
                # 参数校验
                if isinstance(proxy, bytes):
                    proxy = proxy.decode('utf-8')
                # ip:port
                real_proxy = 'http://' + proxy
                try:
                    async with session.get(TEST_API, headers=TEST_HEADERS, proxy=real_proxy, timeout=10) as response:
                        if response.status == 200:
                            # 代理是有效的
                            # 添加到代理池
                            self.__conn.put(proxy)
                            print('有效代理！', proxy)

                except Exception:
                    print('无效代理！', proxy)

        except Exception as e:
            print(e)

    def tester(self):
        print('校验器开始工作！')

        loop = asyncio.get_event_loop()
        # 执行方法
        # 任务列表
        # 从篮子里取出每一个代理，执行上面方法即可
        tasks = [self.test_singer_proxy(x) for x in self.__raw_proxies]
        # 监听
        loop.run_until_complete(asyncio.wait(tasks))


class PoolAdder(object):
    def __init__(self, threshold):
        self.__threshold = threshold
        self.__conn = Reidis_client()
        self.__tester = VailidtyTester()
        self.__crawler = FreeProxyGetter()

    # 停止条件
    def is_over_threshold(self):
        '''
        判断是否超过最大值
        :return: True:超过:添加器停止工作
        '''
        if self.__threshold <= self.__conn.queue_len:
            return True
        return False

    def add_to_queue(self):
        print('添加器开始工作！')
        while True:
            # 超过最大值，停止工作
            if self.is_over_threshold():
                break
            # 1、调用crawler获取代理
            '''
            问题是：现在只能通过对象调用指定方法。
            想调用所有的方法如何实现。
            '''
            # proxies1 = self.__crawler.crawl_66ip()
            # proxies2 = self.__crawler.crawl_ip3366()
            # proxies3 = self.__crawler.crawl_ip8877()
            # proxies3 = self.__crawler.crawl_ip8877()
            # 循环---关键就是我从哪里取哪这个__crawler的这些爬代理的方法名【crawl_66ip，crawl_66ip，crawl_66ip】
            proxies_count = 0
            for crawl_func in self.__crawler.__crawlfuncs__:
                try:
                    proxies = self.__crawler.get_proxies(crawl_func)
                except Exception:
                    continue

                # 2、调用校验器检验并添加
                self.__tester.set_raw_proxies(proxies)
                self.__tester.tester()
                proxies_count += len(proxies)

                if proxies_count == 0:
                    print('代理网站全部失效了！请添加代理的获取方法！')


class Sheduler(object):
    # 从代理池中取出之前爬取的前面的一般，重新校验
    @staticmethod
    def vaild_proxy(cycle=CYCLE_VAILD_TIME):
        conn = Reidis_client()
        tester = VailidtyTester()
        while True:
            print('循环校验器开始工作！')
            count = int(conn.queue_len * 0.5)
            if count == 0:
                print('代理池的数量不足，请添加！')
                time.sleep(cycle)
            proxies = conn.get(count)
            # 校验
            tester.set_raw_proxies(proxies)
            tester.tester()
            time.sleep(cycle)

    @staticmethod
    def check_pool_add(lower_num=LOWER_NUM, upper_num=UPPER_NUM,
                       cycle=CHECK_POOL_CYCLE):
        conn = Reidis_client()
        adder = PoolAdder(upper_num)
        while True:

            if conn.queue_len < lower_num:
                adder.add_to_queue()

            time.sleep(cycle)

    def run(self):
        vaild_procee = Process(target=Sheduler.vaild_proxy)
        check_procee = Process(target=Sheduler.check_pool_add)
        vaild_procee.start()
        check_procee.start()



from setuptools import setup, find_packages

setup(
   name='ss_proxy_pool',
   version='3.1',
   packages=find_packages(),
   author='Example Author',
   author_email='',
   keywords='',
   description='A small example package',
   license='',
   url='',
   include_package_data=True,
   install_requires=[],
)



'''

爬取免费代理ip，返回ip
'''
import requests
from lxml import etree


# 设定一个元类
# 第一步：继承type
class ProxyMetaclass(type):
    def __new__(cls, name, bases, attrs):
        """
        定义一个属性__crawlfuncs__=[]
        __crawlcount__=int---爬取方法的数量
        :param name: 类的名称
        :param bases: 类的继承元组
        :param attrs: 类的属性字典
        :return:
        """
        __crawlfuncs__ = []
        count = 0
        for k, v in attrs.items():
            if 'crawl_' in k:
                __crawlfuncs__.append(k)
                count += 1
        attrs['__crawlfuncs__'] = __crawlfuncs__
        attrs['__crawlcount__'] = count

        # 最终还是由type来创建这个类的。
        # 调用父类的方法来创建
        # 父类type他也是通过—__new__来创建的
        # d
        return type.__new__(cls, name, bases, attrs)
        # return super().__new__(name,bases,attrs)


class FreeProxyGetter(object, metaclass=ProxyMetaclass):
    def __init__(self):
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.106 Safari/537.36',
        }

    # ['crawl_66ip', 'crawl_ip3366', 'crawl_ip8877', 'crawl_ip8899']
    def get_proxies(self, crawl_func):
        # crawl_66ip
        proxies = []
        # self.crawl_ip3366()
        # proxies = eval('self.{}()'.format(crawl_func))
        for proxy in eval('self.{}()'.format(crawl_func)):
            # 设定取得个数
            # if len(proxies)<1000:
            #     proxies.append(proxy)
            # else:
            #     yield proxies
            #     #清空proxies
            proxies.append(proxy)
        return proxies

    def crawl_66ip(self):
        # global headers
        """
        url:http://www.66ip.cn/
        :return:list[proxy]
        """
        proxies = []
        base_url = 'http://www.66ip.cn/%s.html'
        for i in range(1, 20):
            response = requests.get(base_url % i, headers=self.headers)
            html = etree.HTML(response.text)
            ips = html.xpath('//tr[position()>1]/td[1]/text()')
            ports = html.xpath('//tr[position()>1]/td[2]/text()')
            if len(ips) == len(ports) and ips and ports:
                for i, ip in enumerate(ips):
                    port = ports[i]
                    # print(ip,port)
                    proxies.append(ip.strip() + ':' + port.strip())
                    yield ip.strip() + ':' + port.strip()
            # print(proxies)
        # return proxies

    def crawl_xiciIP(self):
        url = 'https://www.xicidaili.com/nn/{}'
        proxies = []
        for i in range(1, 2):
            response = requests.get(url.format(i), headers=self.headers)
            html = etree.HTML(response.text)
            ips = html.xpath('//tr[position()>1]/td[1]/text()')
            ports = html.xpath('//tr[position()>1]/td[2]/text()')
            if len(ips) == len(ports) and ips and ports:
                for i, ip in enumerate(ips):
                    port = ports[i]
                    proxies.append(ip.strip() + ':' + port.strip())
                    # print(proxies)
                    yield ip.strip() + ':' + port.strip()

    def crawl_ipkuai(self):
        '''
       url:http://www.ip3366.net/?stype=1&page=1
       :return:list[proxy]
       '''
        # global headers
        proxies = []
        base_url = 'https://www.kuaidaili.com/free/inha/{}/'
        for i in range(1, 20):
            response = requests.get(base_url.format(i), headers=self.headers)
            html = etree.HTML(response.text)
            ips = html.xpath('//*[@id="list"]/table/tbody/tr/td[1]/text()')
            ports = html.xpath('//*[@id="list"]/table/tbody/tr/td[2]/text()')
            if len(ips) == len(ports) and ips and ports:
                for i, ip in enumerate(ips):
                    port = ports[i]
                    # print(ip,port)
                    proxies.append(ip.strip() + ':' + port.strip())
                    yield ip.strip() + ':' + port.strip()
        # return proxies


if __name__ == '__main__':
    f = FreeProxyGetter()
    # f.crawl_ipkuai()
    attrs_crawl = dir(f)
    crawl_methods = [x for x in attrs_crawl if 'crawl_' in x]
    print(crawl_methods)
    attrs = dir(f)
    print(attrs)
    print(f.__crawlfuncs__)
